# Author:  Radek Marik, RMarik@gmail.com
# Created: April 22, 2017
# Purpose: Pyhop - blocks domain example

# Copyright (c) 2017, Radek Marik, FEE CTU, Prague
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of FEE CTU nor the
#       names of contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from __future__ import print_function
# import os.path

PYHOP_VERSION = 1  # 1.2.2 (official) or 2.0

if PYHOP_VERSION == 2:
    from pyhop import helpers, hop
else:
    import pyhop
    from pyhop import pyhop as plan
    Goal = pyhop.State
    print_goal = pyhop.print_state

Set_C = set

"""The blocks-world operators use three state variables:
- pos[b] = block b's position, which may be 'table', 'hand', or another block.
- clear[b] = False if a block is on b or the hand is holding b, else True.
- holding = name of the block being held, or False if the hand is empty.

A goal is a collection of some (but not necessarily all) of the state variables
and their desired values. Below, both goal1a and goal1b specify c on b, and b
on a. The difference is that goal1a also specifies that a is on table and the
hand is empty.
"""

VERBOSE = 1

# ===================== Support Structures ==============================
print("Define the initial state1: a on b, b on table, c on table")
state1 = pyhop.State('state1')
state1.pos = {'a': 'b', 'b': 'table', 'c': 'table'}
state1.clear = {'c': True, 'b': False, 'a': True}
state1.holding = False
pyhop.print_state(state1)
print('')


# ===================== Support Functions ===============================
def block_is_done(b1, state, goal, done_state):

    if b1 == done_state:
        return True

    if b1 in goal.pos and goal.pos[b1] != state.pos[b1]:
        return False

    if state.pos[b1] == done_state:
        return True

    return block_is_done(state.pos[b1], state, goal, done_state)


def all_blocks(state):
    """Returns all block ids."""
    return state.clear.keys()


def block_status(block_id, state, goal, done_state):
    """
    A helper function used in the methods' preconditions.
    """
    if block_is_done(block_id, state, goal, done_state):
        return 'done'

    elif not state.clear[block_id]:
        return 'inaccessible'

    elif not (block_id in goal.pos) or goal.pos[block_id] == done_state:
        return 'move-to-table'

    elif block_is_done(goal.pos[block_id], state, goal, done_state) and \
            state.clear[goal.pos[block_id]]:
        return 'move-to-block'

    else:
        return 'waiting'


# ===================== Classes =========================================
# ===================== Main Functions ==================================

"""
Operators
"""


def pickup_o(state, x):
    """
    Lift clear block x with the empty arm 
    """
    if state.pos[x] == 'table' and state.clear[x] is True and state.holding is False:
        state.pos[x] = 'hand'
        state.clear[x] = False
        state.holding = x
        return state
    else:
        return False


def unstack_o(state, x, y):
    """
    Pick up clear block x from block y 
    """
    if state.pos[x] == y and y != 'table' and state.clear[x] is True and state.holding is False:
        state.pos[x] = 'hand'
        state.clear[x] = False
        state.holding = x
        state.clear[y] = True
        return state
    else:
        return False


def putdown_o(state, x):
    """
    Place the held block x onto a free space on the table
    """
    if state.pos[x] == 'hand':
        state.pos[x] = 'table'
        state.clear[x] = True
        state.holding = False
        return state
    else:
        return False


def stack_o(state, x, y):
    """
    Place block x using the arm onto clear block y 
    """
    if state.pos[x] == 'hand' and state.clear[y] is True:
        state.pos[x] = y
        state.clear[x] = True
        state.holding = False
        state.clear[y] = False
        return state
    else:
        return False

pyhop.declare_operators(pickup_o, unstack_o, putdown_o, stack_o)

"""
Methods
"""


def move_blocks_m(state, goal):
    """
    This method implements the following block-stacking algorithm:
    If there's a block that can be moved to its final position, then
    do so and call move_blocks recursively. Otherwise, if there's a
    block that needs to be moved and can be moved to the table, then
    do so and call move_blocks recursively. Otherwise, no blocks need
    to be moved.
    """
    for block in all_blocks(state):
        s = block_status(block, state, goal, 'table')
        if s == 'move-to-table':
            return [('move_one', block, 'table'), ('move_blocks', goal)]
        elif s == 'move-to-block':
            return [('move_one', block, goal.pos[block]), ('move_blocks', goal)]
        else:
            continue

    # if we get here, no blocks can be moved to their final locations
    block = pyhop.find_if(lambda x: block_status(x, state, goal, 'table') == 'waiting', all_blocks(state))

    if block is not None:
        return [('move_one', block, 'table'), ('move_blocks', goal)]

    # if we get here, there are no blocks that need moving
    return []


# 'move_block' is an abstract method
# the rest of parameters is one or a list of procedural functions performing variants of decompositions
pyhop.declare_methods('move_blocks', move_blocks_m)


# Method "move_one"
def move_one_block_m(state, block, dest):
    """
    Generate subtasks to get block and put it at dest.
    """
    return [('get', block), ('put', block, dest)]


pyhop.declare_methods('move_one', move_one_block_m)

if 1:
    # Direct method
    # ==============
    # Method "get"
    # ------------
    def get_block_m(state, block):
        """
        Generate either a pickup or an unstack subtask for b1.
        """
        if state.clear[block]:
            if state.pos[block] == 'table':
                return [('pickup_o', block)]
            else:
                return [('unstack_o', block, state.pos[block])]
        else:
            return False

    pyhop.declare_methods('get', get_block_m)

else:
    # Backtracking method
    # =====================
    # Method "get"
    # ------------
    def get_by_unstack_m(state, block):
        """Generate a pickup subtask."""
        if state.clear[block]:
            return [('unstack_task', block)]
        return False


    def get_by_pickup_m(state, block):
        """Generate a pickup subtask."""
        if state.clear[block]:
            return [('pickup_task', block)]
        return False


    pyhop.declare_methods('get', get_by_pickup_m, get_by_unstack_m)

    # Method "pickup_task"
    def pickup_m(state, block):
        """Generate a pickup subtask."""
        if state.clear[block]:
            return [('pickup_o', block)]
        return False

    pyhop.declare_methods('pickup_task', pickup_m)

    # Method "unstack_task"
    def unstack_m(state, block):
        """Generate a pickup subtask."""
        if state.clear[block]:
            return [('unstack_o', block, state.pos[block])]
        return False

    pyhop.declare_methods('unstack_task', unstack_m)


# Method "put"
def put_block_m(state, x, y):
    """
    Generate either a putdown or a stack subtask for block x.
    Block y is x's destination: either the table or another block.
    """
    if state.holding == x:
        if y == 'table':
            return [('putdown_o', x)]
        else:
            return [('stack_o', x, y)]
    else:
        return False


pyhop.declare_methods('put', put_block_m)


# ===================== Resource Registration ===========================
# ===================== Test Functions ==================================

def test_operators_methods():
    # =============================

    print('- these should fail:\n=======================')
    plan(state1, [('pickup_o', 'a')], verbose=VERBOSE)
    plan(state1, [('pickup_o', 'b')], verbose=VERBOSE)

    print('- these should succeed:\n=======================')
    plan(state1, [('pickup_o', 'c')], verbose=VERBOSE)
    plan(state1, [('unstack_o', 'a', 'b')], verbose=VERBOSE)
    plan(state1, [('get', 'a')], verbose=VERBOSE)

    print('- this should fail:\n=======================')
    plan(state1, [('get', 'b')], verbose=VERBOSE)

    print('- this should succeed:\n=======================')
    plan(state1, [('get', 'c')], verbose=VERBOSE)


def test_plan1():
    # ==================
    print("Define the goal goal1a:\n=======================")
    goal1a = Goal('goal1a')
    goal1a.pos = {'c': 'b', 'b': 'a', 'a': 'table'}
    goal1a.clear = {'c': True, 'b': False, 'a': False}
    goal1a.holding = False
    print_goal(goal1a)
    print('')
    plan(state1, [('move_blocks', goal1a)], verbose=VERBOSE)

    # goal1b omits some of the conditions of goal1a,
    # but those conditions will need to be achieved anyway
    print("Define the goal goal1b:\n=======================")
    goal1b = Goal('goal1b')
    goal1b.pos = {'c': 'b', 'b': 'a'}
    print_goal(goal1b)
    print('')
    plan(state1, [('move_blocks', goal1b)], verbose=VERBOSE)


def test_plan2():
    # A bit more complex domain
    print("Define the initial state2:")
    state2 = pyhop.State('state2')
    state2.pos = {'a': 'c', 'b': 'd', 'c': 'table', 'd': 'table'}
    state2.clear = {'a': True, 'c': False, 'b': True, 'd': False}
    state2.holding = False
    pyhop.print_state(state2)
    print('')

    print("Define the goal goal2a:\n=======================")
    goal2a = Goal('goal2a')
    goal2a.pos = {'b': 'c', 'a': 'd', 'c': 'table', 'd': 'table'}
    goal2a.clear = {'a': True, 'c': False, 'b': True, 'd': False}
    goal2a.holding = False
    print_goal(goal2a)
    print('')
    plan(state2, [('move_blocks', goal2a)], verbose=VERBOSE)

    print("Define the goal goal2b:\n=======================")
    goal2b = Goal('goal2b')
    goal2b.pos = {'b': 'c', 'a': 'd'}
    print_goal(goal2b)
    print('')
    plan(state2, [('move_blocks', goal2b)], verbose=VERBOSE)


def test_plan3():
    # ==================
    # An even more complex domain
    state3 = pyhop.State('state3')
    state3.pos = {1: 12, 12: 13, 13: 'table', 11: 10, 10: 5, 5: 4, 4: 14, 14: 15,
                  15: 'table', 9: 8, 8: 7, 7: 6, 6: 'table', 19: 18, 18: 17, 17: 16,
                  16: 3, 3: 2, 2: 'table'}
    state3.clear = {x: False for x in range(1, 20)}
    state3.clear.update({1: True, 11: True, 9: True, 19: True})
    state3.holding = False
    pyhop.print_state(state3)
    print('')

    print("Define the goal goal3:\n=======================")
    goal3 = Goal('goal3')
    goal3.pos = {15: 13, 13: 8, 8: 9, 9: 4, 4: 'table', 12: 2, 2: 3, 3: 16, 16: 11, 11: 7,
                 7: 6, 6: 'table'}
    goal3.clear = {17: True, 15: True, 12: True}
    print_goal(goal3)
    print('')
    print('Plan:')
    a_plan = plan(state3, [('move_blocks', goal3)], verbose=VERBOSE)
    pyhop.print_state(state3)
    for action in a_plan:
        print('action:', action)

# ===================== MAIN ============================================
if __name__ == '__main__':

    if 0:
        test_operators_methods()
    if 1:
        test_plan1()
    if 0:
        test_plan2()
    if 0:
        test_plan3()
